<?php

namespace Drupal\riddle_marketplace;

use Drupal\Component\Serialization\Json;
use Drupal\Core\Cache\CacheBackendInterface;
use Drupal\Core\Config\ConfigFactoryInterface;
use Drupal\Core\Datetime\DrupalDateTime;
use Drupal\Core\Logger\LoggerChannelTrait;
use Drupal\Core\Utility\Error;
use Drupal\riddle_marketplace\Exception\NoApiKeyException;
use GuzzleHttp\Client;

/**
 * Class RiddleFeedService.
 *
 * @package Drupal\riddle_marketplace
 */
class RiddleFeedService implements RiddleFeedServiceInterface {

  use LoggerChannelTrait;

  const API_RIDDLES_LIST = 'https://www.riddle.com/creator/api/v3/riddle/list';

  const API_RIDDLE = 'https://www.riddle.com/creator/api/v3/riddle';

  const API_PROJECTS_LIST = 'https://www.riddle.com/creator/api/v3/project/list';

  /**
   * Cache Service to store Riddle Feed.
   *
   * @var \Drupal\Core\Cache\CacheBackendInterface
   */
  private $cacheService;

  /**
   * Riddle Marketplace Module configuration.
   *
   * @var \Drupal\Core\Config\ImmutableConfig
   */
  private $moduleSettings;

  /**
   * Cache validity period.
   *
   * Mainly used to reduce number of requests to Riddle
   * and to keep fast response for user.
   *
   * - period has to be valid for DrupalDateTime::modify method
   * - period should be less then time required to add new Riddle entry,
   *   so that client after adding entry in Riddle can find it in search here.
   *
   * @var string
   */
  private static $cachePeriod = '30 seconds';

  /**
   * Generic title used for Riddles without defined title.
   *
   * -> Riddle UID will be appended at end of it.
   *
   * @var string
   */
  private $emptyTitlePrefix;

  /**
   * Riddle Feed Service.
   *
   * Constructor.
   *
   * @param \Drupal\Core\Cache\CacheBackendInterface $cacheService
   *   Cache service created for caching of Riddle Feed.
   * @param \Drupal\Core\Config\ConfigFactoryInterface $configService
   *   Configuration Factory.
   */
  public function __construct(CacheBackendInterface $cacheService, ConfigFactoryInterface $configService) {
    $this->cacheService = $cacheService;

    // Fetch and store settings for this module.
    $this->moduleSettings = $configService->get('riddle_marketplace.settings');

    // Set Empty Title Prefix.
    $this->emptyTitlePrefix = $this->moduleSettings->get('riddle_marketplace.empty_title_prefix');
  }

  /**
   * {@inheritdoc}
   */
  public function getFeed() {
    $token = $this->getToken();
    $cacheId = 'riddle_marketplace.feed:' . $token;

    if ($cache = $this->cacheService->get($cacheId)) {
      $feed = $cache->data;
    }
    else {
      $feed = [];
      $projects = $this->moduleSettings->get('riddle_marketplace.projects') ?? [NULL];
      foreach ($projects as $project) {
        $page = 1;
        $riddleResponse = $this->fetchRiddles($page, $project);
        $feed += $this->processRiddlesResponse($riddleResponse);
        do {
          $riddleResponse = $this->fetchRiddles($page);
          $feed += $this->processRiddlesResponse($riddleResponse);
          $page++;
        } while (!empty($riddleResponse['data']));
      }

      $cacheExpire = $this->getCacheExpireTimestamp();
      $this->cacheService->set($cacheId, $feed, $cacheExpire);
    }
    return $feed;
  }

  /**
   * Get Riddle Token from riddle_marketplace settings.
   *
   * @return string
   *   Return defined Riddle Token.
   */
  private function getToken() {
    return $this->moduleSettings->get('riddle_marketplace.token');
  }

  /**
   * Fetches feed from Riddle API.
   *
   * @param int $page
   *   (optional) Page number, default 1.
   *
   * @return array
   *   JSON decoded Riddle API result.
   */
  private function fetchRiddles($page = 1, $project = NULL) {
    $response = FALSE;

    if (!$token = $this->getToken()) {
      throw new NoApiKeyException();
    }

    try {
      $params = [
        'sortBy' => 'created',
        'sortOrder' => 'DESC',
        'page' => $page,
      ];
      if ($project) {
        $params['team'] = $project;
      }

      $options = [
        'headers' =>  [
          'Authorization' => 'Bearer ' . $token,
        ],
        'body' => Json::encode($params),
      ];

      $client = new Client();
      $result = $client->post(static::API_RIDDLES_LIST, $options);
      $response = Json::decode($result->getBody());
    }
    catch (\Exception $e) {
      $logger = $this->getLogger('riddle_marketplace');
      Error::logException($logger, $e);
    }

    return $response;
  }

  /**
   * Gets Riddle from Riddle API.
   *
   * @param string $id
   *   Riddle id.
   *
   * @return array
   *   JSON decoded Riddle API result.
   */
  public function getRiddle($id) {
    $response = FALSE;

    try {
      $options = [
        'headers' =>  [
          'Authorization' => 'Bearer ' . $this->getToken(),
        ],
      ];

      $client = new Client();
      $result = $client->get(static::API_RIDDLE . '/' . $id, $options);
      $response = Json::decode($result->getBody());
    }
    catch (\Exception $e) {
      $logger = $this->getLogger('riddle_marketplace');
      Error::logException($logger, $e);
    }

    return $this->processRiddleResponse($response['data'] ?? []);
  }

  /**
   * Gets projects from Riddle API.
   *
   * @return array
   *   Projects ID-name array.
   */
  public function getProjects() {
    $response = FALSE;

    try {
      $client = new Client();
      $options = [
        'headers' =>  [
          'Authorization' => 'Bearer ' . $this->getToken(),
        ],
      ];

      $result = $client->get(static::API_PROJECTS_LIST, $options);
      $response = Json::decode($result->getBody());
    }
    catch (\Exception $e) {
      $logger = $this->getLogger('riddle_marketplace');
      Error::logException($logger, $e);
    }

    if (empty($response['data']['items'])) {
      return [];
    }

    return $response;
  }

  /**
   * Process riddles response from Riddle API.
   *
   * Response is in JSON format. Method will return only relevant data
   * for internal feed cached storage.
   * - currently: uid, title.
   *
   * @param array|null $riddleResponse
   *   JSON or NULL as Riddle API Result.
   *
   * @return array
   *   Filtered Riddle Feed with params relevant for the module.
   */
  private function processRiddlesResponse($riddleResponse) {
    $feed = [];

    if (!empty($riddleResponse['data']) && is_array($riddleResponse['data'])) {
      foreach ($riddleResponse['data'] as $riddleEntry) {
        if ($processed = $this->processRiddleResponse($riddleEntry)) {
          $feed[$riddleEntry['UUID']] = $processed;
        }
      }
    }

    return $feed;
  }

  /**
   * Process riddle response from Riddle API.
   *
   * @param array|null $riddleEntry
   *   JSON or NULL as Riddle API Result.
   *
   * @return array
   *   Filtered Riddle Feed entry with params relevant for the module.
   */
  protected function processRiddleResponse($riddleEntry) {
    if (!$this->isValidRiddleFeedEntry($riddleEntry)) {
      // Skip invalid riddle feed entries.
      return NULL;
    }

    return [
      'title' => $this->getRiddleTitle($riddleEntry),
      'id' => $riddleEntry['UUID'],
      // All riddles in response are published atm.
      'status' => 1,
      'image' => $riddleEntry['image'] ?? NULL,
    ];
  }

  /**
   * Validation Riddle Feed Entry.
   *
   * @param array|null $riddleEntry
   *   Single Riddle Feed Entry.
   *
   * @return bool
   *   Result of validation.
   */
  private function isValidRiddleFeedEntry($riddleEntry) {
    if (
      empty($riddleEntry) || !is_array($riddleEntry) || empty($riddleEntry['UUID'])
    ) {
      return FALSE;
    }

    return TRUE;
  }

  /**
   * Returns Riddle Title.
   *
   * From feed entry return Title
   * - in case title is not defined use generic name.
   *
   * @param array $riddleEntry
   *   Single Riddle Feed Entry.
   *
   * @return string
   *   Riddle element title.
   */
  private function getRiddleTitle(array $riddleEntry) {
    if (!empty($riddleEntry['title'])) {
      return html_entity_decode($riddleEntry['title']);
    }
    return $this->emptyTitlePrefix . $riddleEntry['id'];
  }

  /**
   * Get cache validity end timestamp.
   *
   * @return int
   *   Timestamp of expire time for Cache.
   */
  private function getCacheExpireTimestamp() {
    /* @var \DateTime $date */
    $date = new DrupalDateTime();
    $date->modify('+' . static::$cachePeriod);

    return $date->getTimestamp();
  }

}
